import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

class Node {

    public long index;
    public long weight;
    public Node parent;

    public long v0;
    public long v1;
    public long v2;
    public long v3;
    public long v00;
    public long v02;

}

public class DRDTreeSolver {

    private static final String LOG_LEVEL = "DEBUG";

    private static void log(String input, String currentLevel) {
        if (LOG_LEVEL.equals("DEBUG")) {
            System.out.println(input);
        } else if (currentLevel.equals(LOG_LEVEL)) {
            System.out.println(input);
        }
    }

    public static long min(long a, long b, long c, long d) {
        return Math.min(Math.min(a, b), Math.min(c, d));
    }

    public static long min(long a, long b, long c) {
        return Math.min(Math.min(a, b), c);
    }

    public static long sum(long a, long b) {
        if (a == Long.MAX_VALUE || b == Long.MAX_VALUE) {
            return Long.MAX_VALUE;
        } else {
            return a + b;
        }
    }

    public static long sum(long a, long b, long c) {
        return sum(sum(a, b), c);
    }

    public static void main(String[] args) throws IOException {

        log("START", "INFO");

        var INPUT_FOLDER = "drdtree/input";

//      load all input graphs in resources/<input_folder>
//      input graph format is
//      first line is number of nodes (n) and number of edges (m), example "100 100"
//      next m lines are edges, defined with start and end index of node and starting with letter "e", example "e 1 2"
//      last line is weight of each node separated by space, starting with lowest index node, example ("2 4 6 1 23 4") for graph with 6 nodes
        var files = Paths.get("resources", INPUT_FOLDER).toFile().listFiles();
        assert files != null;
        for (var file : files) {
            if (file.isDirectory()) {
                continue;
            }
            log("PROCESSING " + file.getName(), "INFO");

            List<String> lines = Files.readAllLines(file.toPath()).stream().filter(line -> line.trim().length() > 0).collect(Collectors.toList());

            int verticesNum = -1;
            int edgesNum = -1;

            String infoLine = lines.get(0);
            String[] parts = infoLine.split("\\s+");
            verticesNum = Integer.parseInt(parts[0]);
            edgesNum = Integer.parseInt(parts[1]);

            var weightLine = lines.get(edgesNum + 1);
            var weights = Arrays.stream(weightLine.split("\\s+")).map(Integer::parseInt).collect(Collectors.toList());

            List<Node> tree = new ArrayList<>();
            for (int verticesIndex = 0; verticesIndex < verticesNum; verticesIndex++) {
                Node node = new Node();
                node.index = verticesIndex;
                node.weight = weights.get(verticesIndex);
                node.v0 = Long.MAX_VALUE;
                node.v1 = Long.MAX_VALUE;
                node.v2 = 2 * node.weight;
                node.v3 = 3 * node.weight;
                node.v00 = 0;
                node.v02 = Long.MAX_VALUE;
                tree.add(node);
            }

            lines.subList(1, edgesNum + 1).forEach(edgeLine -> {
                String[] edgeParts = edgeLine.split("\\s+");
                int parentIndex = Integer.parseInt(edgeParts[1]);
                int childIndex = Integer.parseInt(edgeParts[2]);

                Node child = tree.get(childIndex);
                child.parent = tree.get(parentIndex);
            });

            for (int i = verticesNum - 1; i > 0; i--) {
                var node = tree.get(i);

                var parent = node.parent;

                parent.v0 = min(sum(parent.v0, node.v0), sum(parent.v0, node.v1), sum(parent.v02, node.v2), sum(parent.v00, node.v3));

                parent.v1 = min(sum(parent.v1, node.v0), sum(parent.v1, node.v1), sum(parent.v00, node.v2, parent.weight), sum(parent.v00, node.v3, parent.weight));

                parent.v2 = min(sum(parent.v2, node.v02), sum(parent.v2, node.v00, node.weight), sum(parent.v2, node.v2), sum(parent.v2, node.v3));

                parent.v3 = min(sum(parent.v3, node.v00), sum(parent.v3, node.v2), sum(parent.v3, node.v3));

                parent.v02 = min(sum(parent.v02, node.v0), sum(parent.v02, node.v1), sum(parent.v00, node.v2), sum(parent.v00, node.v3));

                parent.v00 = sum(parent.v00, min(node.v0, node.v1, node.v2, node.v3));
            }

            var root = tree.get(0);
            var result = min(root.v0, root.v1, root.v2, root.v3);

            // output is min total cost DRD
            log("RESULT " + result, "INFO");
            log("DONE", "INFO");
        }

    }
}
